// Internal representation of CRP file fragments
// (C) 2003 Frank Paehlke <frank@paehlke.de>
// $Id: crpfile.cc,v 1.7 2004/06/12 09:18:50 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "globals.h"
#include "crpfile.h"
#include "smooth.h"

#include <cmath>
#include <cstdlib>
#include <iomanip>
#include <sstream>

// copy constructor

CRPFile::CRPFile (const CRPFile& crp)
{
    *this = crp;
}

// clear all data values

void CRPFile::clear ()
{
    time.clear();
    distance.clear();
    height.clear();
    hrate.clear();
    temperature.clear();
    comment.clear();
    cdistance.clear();
    cheight.clear();
    speed.clear();
    vspeed.clear();
    slope.clear();
    nrows = 0;
    hdiff = hdiff_up = 0.0;
    length = length_up = 0.0;
    avgslope = avgslope_up = maxslope = 0.0;
    title = "";
    time_t t; // set tourdate to current date
    std::time(&t);
    localtime_r(&t, &tourdate);
}

// read from file

istream& operator >> (istream& is, CRPFile& crp)
    throw (CRPFile::MalformedInputException)
{
    string line; // one input line
    int ival; // integer value read
    double dval; // double value read
    string sval; // string value read

    crp.clear();
    if (!(getline(is,line)))
	throw CRPFile::MalformedInputException();
    if (!line.substr(0,9).compare("HRMProfil")) {
	// input includes complete CRP header
	// we can safely ignore the 2nd header line since we do not need
	// any of the contained data
	if (!(getline(is,line)))
	    throw CRPFile::MalformedInputException();
	// read first real input line into buffer, return if empty
	if (!getline(is,line) || !(is >> ws))
	    return is;
    }
    // read from input file until last line
    do {
	if (!line.substr(0,3).compare("***")) {
	    // end of data rows reached; the input seems to contain a
	    // complete CRP trailer: read it
	    // ignore first line
	    if (!getline(is,line))
		throw CRPFile::MalformedInputException();
	    // first line starts with date/time
	    if (!(is >> sval))
		throw CRPFile::MalformedInputException();
	    string date = sval;
	    if (!(is >> sval))
		throw CRPFile::MalformedInputException();
	    date += ' '; date += sval;
	    if (!(strptime (date.c_str(), "%d.%m.%Y %H:%M:%S",
			    &crp.tourdate)))
		throw CRPFile::MalformedInputException();
	    // ignore duration, category, difficulty
	    for (int i=0; i<3; i++) {
		if (!(is >> sval))
		    throw CRPFile::MalformedInputException();
	    }
	    // read title
	    if (!(is >> ws) || !getline(is, sval))
		throw CRPFile::MalformedInputException();
	    // remove trailing CR
	    if (!sval.empty() && (sval[sval.size()-1] == '\r'))
		sval = sval.substr(0, sval.size()-1); 
	    crp.title = sval;
	    // the rest of the file can be ignored
	    break;
	}
	istringstream iss(line);
	// heart rate
	if (!(iss >> ival))
	    throw CRPFile::MalformedInputException();
	crp.hrate.push_back(ival);
	// speed (ignored)
	if (!(iss >> ival)) 
	    throw CRPFile::MalformedInputException();
	// distance 
	if (!(iss >> ival)) 
	    throw CRPFile::MalformedInputException();
	crp.distance.push_back(ival/100.0);
	if ((crp.distance.size() > 1) &&
	    (crp.distance.back() == crp.distance[crp.distance.size()-2])) {
	    cerr << "Warning: standstill at km " <<
		crp.distance.back() << endl;
	}
	// height 
	if (!(iss >> dval)) 
	    throw CRPFile::MalformedInputException();
	crp.height.push_back(dval);
	// ignore color + symbol
	if (!(iss >> ival))
	    throw CRPFile::MalformedInputException();
	if (!(iss >> ival))
	    throw CRPFile::MalformedInputException();
	// temperature (format: 99,9)
	{ 
	    if (!(iss >> ival)) 
		throw CRPFile::MalformedInputException();
	    double temp = ival;
	    char comma;
	    if (!(iss >> comma) || !(iss >>ival))
		throw CRPFile::MalformedInputException();
	    temp += ival/10.0;
	    crp.temperature.push_back(temp);
	}
	// time
	if (!(iss >> ws) || !(getline (iss, sval, '\t')))
	    throw CRPFile::MalformedInputException();
	crp.time.push_back(sval);
	// comment (rest of line)
	if (!(getline (iss, sval)))
	    throw CRPFile::MalformedInputException();
	// remove trailing CR
	if (!sval.empty() && (sval[sval.size()-1] == '\r'))
	    sval = sval.substr(0, sval.size()-1); 
	crp.comment.push_back(sval);
    } while (getline(is, line) && (is >> ws));
    crp.calculate(); // calculate derived data rows and aggregated values
    return is;
}

// calculate derived data rows and aggregated values

void CRPFile::calculate()
{
    int i;
    nrows = time.size(); // number of table rows

    // calculate corrected distance
    
    // 1st step: smoothing
    if (sm == 0) {
	cdistance = distance;
    }
    else {
	if (globals.verbose)
	    cerr << "smoothing distance values" << endl;
	sm->smooth_vector (distance, cdistance, 0.005);
    }
    // 2nd step: scaling
    if (globals.verbose)
	cerr << "scaling distance values" << endl;
    double startdistance = cdistance[0];
    for (i=0; i<nrows; i++) {
	cdistance[i] = globals.lscale * (cdistance[i] - startdistance);
    }

    // calculate corrected height

    // 1st step: smoothing
    if (sm == 0) {
	cheight = height;
    }
    else {
	if (globals.verbose)
	    cerr << "smoothing height values" << endl;
	sm->smooth_vector (height, cheight, 0.5);
    }
    // 2nd step: scaling
    if (globals.verbose)
	cerr << "scaling height values" << endl;
    double oldstartheight = cheight[0];
    double startheight = oldstartheight + globals.hoffset;
    if (globals.use_startheight)
	startheight = globals.startheight;
    double oldendheight = cheight[nrows-1];
    double endheight = oldendheight + globals.hoffset;
    if (globals.use_endheight)
	endheight = globals.endheight;
    // determine height scaling factor
    double hscale;
    if (fabs(oldendheight - oldstartheight) >= 0.5) {
	hscale 	= (endheight-startheight) / (oldendheight-oldstartheight);
    }
    else {
	if (fabs(endheight - startheight) >= 0.5) {
	    cerr << "cannot scale file with zero height difference" << endl;
	    exit (1);
	}
	else
	    hscale = 1;
    }
    // do actual scaling
    for (i=0; i<nrows; i++) {
	cheight[i] = startheight + hscale * (cheight[i] - oldstartheight);
    }

    // calculate speed (assuming 20sec intervals)

    if (globals.verbose)
	cerr << "calculating speed values" << endl;
    speed.resize (nrows);
    speed[0] = 0;
    for (i=1; i<nrows; i++) {
	speed[i] = (cdistance[i] - cdistance[i-1]) * 3600.0 / 20.0;
    }

    // calculate slope

    if (globals.verbose)
	cerr << "calculating slope values" << endl;
    slope.resize (nrows);
    slope[0] = 0;
    for (i=1; i<nrows; i++) {
	slope[i] = (cheight[i] - cheight[i-1])
	    / (cdistance[i] - cdistance[i-1]) / 10.0;
    }

    // calculate vspeed from speed and slope

    if (globals.verbose)
	cerr << "calculating vspeed values" << endl;
    vspeed.resize (nrows);
    vspeed[0] = 0;
    for (i=1; i<nrows; i++) {
	vspeed[i] = speed[i] * slope[i] * 10.0;
    }

    // total length and height difference

    length = cdistance[nrows-1] - cdistance[0];
    hdiff = round(cheight[nrows-1] - cheight[0]);
    
    // lenght and height difference of ascent/descent only

    length_up = 0;
    length_down = 0;
    hdiff_up = 0;
    hdiff_down = 0;
    for (i=1; i<nrows; i++) {
	if (cheight[i] > cheight[i-1]) {
	    length_up += (cdistance[i] - cdistance[i-1]);
	    hdiff_up += (cheight[i] - cheight[i-1]);
	}
	else if (cheight[i] < cheight[i-1]) {
	    length_down += (cdistance[i] - cdistance[i-1]);
	    hdiff_down += (cheight[i-1] - cheight[i]);
	}
    }
    
    // average slope

    avgslope = hdiff / length / 10.0;
    avgslope_up = (length_up > 0)
	? hdiff_up / length_up / 10.0
	: 0.0;
    avgslope_down = (length_down > 0)
	? hdiff_down / length_down / 10.0
	: 0.0;


    // maximum/minimum slope

    maxslope = minslope = slope[1];
    for (i=2; i<nrows; i++) {
	if (slope[i] > maxslope)
	    maxslope = slope[i];
	if (slope[i] < minslope)
	    minslope = slope[i];
    }
}

// TODO: write contents to file

ostream& operator << (istream& os, const CRPFile& crp)
{
    cerr << "Error: CRP output not yet implemented" << endl;
    exit(1);
}

// dump file contents as CSV

void CRPFile::writeCSV (ostream& os) const
{
    char datestr[12];
    strftime (datestr, 12, "%Y-%m-%d", &tourdate);
    os << "Title\t\"" << title << '\"' << endl
       << "Date\t" << datestr << endl
       << "Length [km]\t" << length << endl
       << "Length ascent [km]\t" << length_up << endl
       << "Length descent [km]\t" << length_down << endl
       << "Hdiff [m]\t" << hdiff << endl
       << "Hdiff ascent [m]\t" << hdiff_up << endl
       << "Hdiff descent [m]\t" << hdiff_down << endl
       << "Avg. slope [%]\t" << avgslope << endl
       << "Avg. slope ascent [%]\t" << avgslope_up << endl
       << "Avg. slope descent [%]\t" << avgslope_down << endl
       << "Max. slope [%]\t" << maxslope << endl
       << "Min. slope [%]\t" << minslope << endl
       << "time\t"
       << "orig. distance [km]\t"
       << "orig. height [m]\t"
       << "heart rate [1/min]\t"
       << "temperature [C]\t"
       << "corr. distance [km]\t"
       << "corr. height [m]\t"
       << "speed [km/h]\t"
       << "vert. speed [m/h]\t"
       << "slope [%]\t"
       << "comment" << endl;
    for (unsigned int i=0; i < time.size(); i++) {
	os << time[i] << '\t' << distance[i] << '\t' 
	   << height[i] << '\t' << hrate[i] << '\t' 
	   << temperature[i] << '\t' << cdistance[i] << '\t' 
	   << cheight[i] << '\t' << speed [i] << '\t'
	   << vspeed [i] << '\t' << slope[i] << '\t'
	   << comment[i] << endl;
    }
}

// dump meta data as XML

void CRPFile::writeXML (ostream& os) const
{
    // TODO: make ID, region, author, title and date
    // configurable as parameters

    string id = "";
    string region = "";
    string subregion = "";
    string author = "";
    
    // XML preamble
    os << "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
       << "<!DOCTYPE profile SYSTEM \"profile-0.3.dtd\">\n\n";

    // profile header
    os << "<profile id=\"" << id << "\">\n";

    // meta data
    os << "<meta>\n"
       << "<dest>" << comment[nrows-1] << "</dest>\n"
       << "<route>" << comment[0] << "</route>\n"
       << "<region>" << region << "</region>\n"
       << "<subregion>" << subregion << "</subregion>\n"
       << "<author>" << author << "</author>\n";

    // tour date
    char datestr[12];
    strftime (datestr, 12, "%Y-%m-%d", &tourdate);
    os << "<date>" << datestr << "</date>\n"
       << "</meta>\n";

    // title
    os << "<title>" << title << "</title>\n";

    // start, finish 
    int startheight = (int)round(cheight[0]);
    int finishheight = (int)round(cheight[nrows-1]);
    os << "<start><height>" << startheight << "</height>"
       << "<p>" << comment[0] << "</p></start>\n"
       << "<finish><height>" << finishheight << "</height>"
       << "<p>" << comment[nrows-1] << "</p></finish>\n";
	
    // should we include separate entries for ascent/descent only?
    // this makes sense only if the ascent/descent is non-zero,
    // but shorter than the total length

    bool ascent = (length_up > 0) && (length_up < (length - 1e-5));
    bool descent = (length_down > 0) && (length_down < (length - 1e-5));

    os << fixed << setprecision(1) << "<length>"
       << "<total>" << length << "</total>";
    if (ascent) {
	os << "<up>" << length_up << "</up>";
    }
    if (descent) {
	os << "<down>" << length_down << "</down>";
    }
    os << "</length>\n";
    
    // height difference
    
    os << setprecision(0) << "<hdiff>"
       << "<total>" << hdiff << "</total>";
    if (ascent) {
	os << "<up>" << hdiff_up << "</up>";
    }
    if (descent) {
	os << "<down>" << hdiff_down << "</down>";
    }
    os << "</hdiff>\n";
    
    // slope

    os << setprecision(1) << "<slope>" 
       << "<total>" << avgslope << "</total>";
    if (ascent) {
	os << "<up>" << avgslope_up << "</up>";
    }
    if (descent) {
	os << "<down>" << avgslope_down << "</down>";
    }
    os << setprecision(0) << "<max>" << maxslope << "</max>";
    os << "</slope>\n";

    // important points (other than start and finish)

    for (int i=1; i<nrows-1; i++) {
	if (!comment[i].empty()) {
	    int pheight = 5 * (int)round(cheight[i]/5.0);
	    os << setprecision(1)
	       << "<point><pos>" << cdistance[i] << "</pos>"
	       << "<height>" << pheight << "</height>"
	       << "<p>" << comment[i] << "</p></point>\n";
	}
    }

    // category, graph, description (empty)

    os << "<categ></categ>\n"
       << "<graph url=\"\" svg=\"\" pdf=\"\"/>\n"
       << "<description>\n</description>\n";
    
    // end of profile
    
    os << "</profile>\n";
}
