#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-

# Create an SVG graph from a CSV file with height profile data.
# The CSV file must contain the distance and height values in
# the 6th and 7th column, respectively.
# CSV files which fulfill this condition are created by crp2csv.

# Author: Frank Paehlke
# $Id: svggraph.py,v 1.4 2004/09/29 19:16:25 frank Exp $

import sys
import math
from optparse import OptionParser
from xml.dom.minidom import getDOMImplementation
import string
import codecs
import csv
import array
import os
import tempfile
import Tkinter
import crputils_version
import svgobject

def main():
    global options
    # parse command line arguments
    parser = OptionParser(usage="%prog [options] [csvfile [outfile]]",
                          version=crputils_version.version)
    parser.add_option ("-v", "--verbose", action="store_true",
                       dest="verbose", default=False,
                       help="turn on debugging output")
    parser.add_option ("-t", "--title", dest="title", default="",
                       help="set title text")
    parser.add_option ("-c", "--copyright", dest="copyright", default="",
                       help="set copyright text")
    parser.add_option ("-f", "--format", dest="format", default="svg",
                       help="output format: svg (default), "+
                       "png (requires ImageMagick), "+
                       "pdf (requires batik-rasterizer), screen")
    parser.add_option ("", "--dscale", dest="dscale", default="",
                       metavar="MIN,MAX",
                       help="set scale of distance axis manually" )
    parser.add_option ("", "--hscale", dest="hscale", default="",
                       metavar="MIN,MAX",
                       help="set scale of height axis manually")
    parser.add_option ("", "--sscale", dest="sscale", default="",
                       metavar="MIN,MAX",
                       help="set scale of slope axis manually")
    parser.add_option ("", "--encoding", dest="encoding", default="",
                       help="set character encoding of CSV file (default: utf-8)")
    (options, args) = parser.parse_args()

    # do some checks
    options.format = string.lower(options.format)
    if options.format not in ['svg','png','pdf','screen']:
        print >> sys.stderr, "unsupported format:", options.format
        sys.exit(1)
    if len(args) > 2:
        print >> sys.stderr, "Ignoring extra argument(s):", args[2:]

    # read input/output file name
    if len(args) >= 1:
        ifilename = args[0]
    else:
        ifilename = "-"
    if len(args) >= 2:
        ofilename = args[1]
    else:
        i = string.rfind(ifilename,'.')
        if i < 0:
            basename = ifilename
        else:
            basename = ifilename[0:i]
        ofilename = basename+'.'+options.format
    if options.verbose:
        print >> sys.stderr, "Input file:", ifilename
        print >> sys.stderr, "Output file:", ofilename
        print >> sys.stderr, "Output format:", options.format

    # read input file
    if ifilename == '-':
        ifile = sys.stdin
    else:
        ifile = open(ifilename)
    dist,height = readFile(ifile)
    if (ifile != sys.stdin):
        ifile.close()

    # decode title/copyright to UTF-8
    if options.encoding == "":
        options.encoding = "utf-8"
    options.title = options.title.decode (options.encoding).encode ("utf-8")
    options.copyright = options.copyright.decode (options.encoding).encode ("utf-8")

    if options.verbose:
        print >> sys.stderr, "Title:", options.title
        print >> sys.stderr, "Copyright text:", options.copyright

    if options.format == "screen":
        # create SVG
        svg = svgobject.SVG (640, 300, unit="px")
        drawGraph (svg, dist, height)
        # display on screen
        root = Tkinter.Tk()
        app = svgobject.SVGView (root, svg)
        app.mainloop()
        root.destroy()

    if options.format == "svg":
        # create SVG
        svg = svgobject.SVG (640, 300, unit="pt")
        drawGraph (svg, dist, height)
        # write output file
        if ofilename == '-':
            ofile = codecs.getwriter("utf-8")(sys.stdout)
        else:
            ofile = codecs.open(ofilename, "wb", "utf-8")
        dom = svg.toDOM(getDOMImplementation())
        dom.writexml (ofile, addindent=" ", newl="\n", encoding="utf-8")
        if ofile != sys.stdout:
            ofile.close()

    elif options.format == "png":
        # create SVG
        svg = svgobject.SVG (640, 300, unit="pt")
        drawGraph (svg, dist, height)
        # create color map
        colormap = create_web_color_map ()
        if options.verbose:
            print >> sys.stderr, "Temporary file for color map:", colormap
        pipe = os.popen("convert -map " + colormap
                        + " +dither -depth 8 -quality 90"
                        + " - " + ofilename, "w");
        dom = svg.toDOM(getDOMImplementation())
        dom.writexml (codecs.getwriter("utf-8")(pipe), encoding="utf-8")
        pipe.close()
        os.remove(colormap)

    elif options.format == "pdf":
        # create A4 with A4 paper size
        svg = svgobject.SVG (297, 210, unit="mm", view_box=(-51.25, -112.5, 742.5, 525))
        drawGraph (svg, dist, height)
        tfd,tfname = tempfile.mkstemp(".svg")
        if options.verbose:
            print >> sys.stderr, "Temporary file:", tfname
        ofile = codecs.getwriter("utf-8")(os.fdopen (tfd,'w'))
        dom = svg.toDOM(getDOMImplementation())
        dom.writexml (ofile, encoding="utf-8")
        ofile.close()
        os.spawnlp(os.P_WAIT, 'batik-rasterizer', 'batik-rasterizer',
                   '-m', 'application/pdf', '-d', ofilename, tfname )
        os.remove(tfname)

# read input data from file f
# returns: dist[], height[]

def readFile(f):
    dist = array.array('d')
    height = array.array('d')
    reader = csv.reader(f,delimiter='\t')
    for line in reader:
        # does the line contain a title?
        token = line[0].strip().lower()
        if token == "title":
            if options.title == "":
                options.title = line[1]
        elif token == "copyright":
            if options.copyright == "":
                options.copyright = line[1]
        elif token == "dscale":
            if options.dscale == "":
                options.dscale = line[1].strip()
        elif token == "hscale":
            if options.hscale == "":
                options.hscale = line[1].strip()
        elif token == "sscale":
            if options.sscale == "":
                options.sscale = line[1].strip()
        elif token == "encoding":
            if options.encoding == "":
                options.encoding = line[1].strip()
        elif line[0][0] in "0123456789.+-":
            # read distance and height
            dist.append(float(line[5]))
            height.append(float(line[6]))
    if options.verbose:
        print >> sys.stderr, len(dist), "data rows read"
    return dist,height

def drawGraph(svg, dist, height):

    root = svgobject.Group (svg, style=svgobject.Style (font_family="Arial,Helvetica,sans-serif", font_size="14"))

    # calculate slope
    slope = [ (b-a)/(d-c)/10.0 for a,b,c,d in
              zip(height[0:-1], height[1:], dist[0:-1], dist[1:]) ]
    # calculate scales
    d0,d1,dIntv = calcXScale(dist)
    h0,h1,hIntv,s0,s1,sIntv = calcYScale(height,slope)
    dTics = int((d1-d0)/dIntv)
    hTics = (h1-h0)/hIntv
    sTics = (s1-s0)/sIntv
    # background
    svgobject.Rect (root, 0, 0, 639, 299,
                    style=svgobject.Style (fill="#ffffff", stroke="#000000", stroke_width="1"))

    # title
    if options.title != "":
        svgobject.Text (root, 320, 30, options.title,
                        style=svgobject.Style (text_anchor="middle", font_weight="bold", font_size="16"))
    # copyright
    if options.copyright != "":
        svgobject.Text (root, 636, 296, options.copyright,
                        style=svgobject.Style (text_anchor="end", font_size="10", fill="#999999"))

    # canvas
    canvas = svgobject.Group (root, translate=(80,50))
    svgobject.Rect (canvas, 0, 0, 500, 200, style=svgobject.Style(fill="#cccccc"))

    # X minor grid
    if dTics <= 12:
        g = svgobject.Group (canvas, style=svgobject.Style (stroke="#999999",stroke_width=1))
        for i in range (dTics):
            d = d0 + (i+0.5)*dIntv;
            x = int(round(scale(d, d0, d1, 0, 500)))
            svgobject.Line (g, x, 0, x, 200)

    # Y minor grid
    if sTics <= 7:
        g = svgobject.Group (canvas, style=svgobject.Style (stroke="#999999",stroke_width=1))
        for i in range (sTics):
            s = s0 + (i+0.5)*sIntv;
            y = int(round(scale(s, s0, s1, 200, 0)))
            svgobject.Line (g, 0, y, 500, y)

    # X grid
    g = svgobject.Group (canvas, style=svgobject.Style (stroke="#000000",stroke_width=1))
    for i in range (dTics+1):
        d = d0 + i*dIntv;
        x = int(round(scale(d, d0, d1, 0, 500)))
        svgobject.Line (g, x, 0, x, 205)

    # X labels
    g = svgobject.Group (canvas, style=svgobject.Style (text_anchor="middle"))
    svgobject.Text (g, 250, 240, "Strecke [km]")
    for i in range (dTics+1):
        d = d0 + i*dIntv;
        x = int(round(500 * float(d-d0) / float(d1-d0)))
        svgobject.Text (g, x, 220, str(round(d,1)))

    # Y grid
    g = svgobject.Group (canvas, style=svgobject.Style (stroke="#000000",stroke_width=1))
    for i in range (hTics+1):
        h = h0 + i*hIntv;
        s = s0 + i*sIntv;
        y = round(scale(h, h0, h1, 200, 0))
        if (s == 0) or (h == 0): # zero axis
            svgobject.Line (g, -5, y, 505, y, style=svgobject.Style(stroke_width=2))
        else:
            svgobject.Line (g, -5, y, 505, y)

    # Y labels (primary)
    g = svgobject.Group (canvas, translate=(-55,100), rotate=-90)
    svgobject.Text (g, 0, 0, u"blau: Hhe [m]", style=svgobject.Style(text_anchor="middle"))
    g = svgobject.Group (canvas, style=svgobject.Style(text_anchor="end"))
    for i in range (hTics+1):
        h = h0 + i*hIntv;
        y = round(scale(h, h0, h1, 200, 0))
        svgobject.Text (g, -10, y+5, str(int(round(h)))) # TODO: dy="0.5ex"

    # Y labels (secondary)
    g = svgobject.Group (canvas, translate=(545,100), rotate=-90)
    svgobject.Text (g, 0, 0, "rot: Steigung [%]", style=svgobject.Style(text_anchor="middle"))
    g = svgobject.Group (canvas, style=svgobject.Style(text_anchor="start"))
    for i in range (sTics+1):
        s = s0 + i*sIntv
        y = round(scale(s, s0, s1, 200, 0))
        svgobject.Text (g, 510, y+5, str(int(round(s)))) # TODO: dy="0.5ex"

    # slope curve (first, i.e. behind height curve)
    dist2 = map((lambda x,y: (x+y)/2), dist[0:-1], dist[1:])
    drawCurve (canvas, dist2,d0,d1,slope,s0,s1,"#ff0000")

    # height curve (second, i.e. before slope curve)
    drawCurve (canvas, dist,d0,d1,height,h0,h1,"#0000ff")

# calculate scaling parameters X axis
# returns: dBottom, dTop, dIntv

def calcXScale (dist):
    if options.dscale != "":
        (dMin, dMax) = string.split(options.dscale, ',')
        dMin = float(dMin)
        dMax = float(dMax)
    else:
        dMin = min(dist)
        dMax = max(dist)
    if options.verbose:
        print "D min/max:", dMin, dMax

    # determine scale
    for dIntv in [0.1, 0.2, 0.5, 1, 2, 4, 5]:
        # round up/down
        dBottom = dIntv * math.floor(dMin/dIntv)
        dTop = dIntv * math.ceil(dMax/dIntv)
        dTics = int(round((dTop-dBottom) / dIntv))
        if (dTics <= 15):
            break # we have found a suitable scale
    if options.verbose:
        print "D scale:", dBottom, dTop, dIntv, dTics
    return dBottom, dTop, dIntv

# calculate scaling parameters for primary and secondary Y axis
# returns: hBottom, hTop, hIntv, sBottom, sTop, sIntv

def calcYScale (height,slope):
    if options.hscale != "":
        (hMin, hMax) = map(float, string.split(options.hscale, ','))
    else:
        hMin = min(height)
        hMax = max(height)
    if options.sscale != "":
        (sMin, sMax) = map(float, string.split(options.sscale, ','))
    else:
        # always include zero axis into slope scale
        sMin = min(slope+[0])
        sMax = max(slope+[0])
    # determine scale for primary axis only
    for hIntv in [5, 10, 20, 50, 100, 200, 400]:
        # round up/down
        hBottom = hIntv*int(math.floor(hMin/hIntv))
        hTop = hIntv*int(math.ceil(hMax/hIntv))
        # hBottom and hTop are integer!
        hTics = (hTop-hBottom) / hIntv
        if (hTics <= 8):
            break # we have found a suitable scale
    if options.verbose:
        print "original H scale:", hBottom, hTop, hIntv, hTics

    # determine scale for secondary axis only
    for sIntv in [1, 2, 4]:
        # round up/down
        sBottom = sIntv*int(math.floor(sMin/sIntv))
        sTop = sIntv*int(math.ceil(sMax/sIntv))
        # sBottom and sTop are integer!
        sTics = (sTop-sBottom) / sIntv
        if (sTics <= 8):
            break # we have found a suitable scale
    if options.verbose:
        print "original S scale:", sBottom, sTop, sIntv, sTics

    # adjust scales to match each other
    if (sTics > hTics):
        # adjust height scale
        d = sTics - hTics
        hBottom -= hIntv * (d/2)
        hTop += hIntv * (d/2)
        if d%2 == 1: # odd number
            if (hTop-hMax < hMin-hBottom):
                hTop += hIntv
            else:
                hBottom -= hIntv
        hTics = sTics
    elif (hTics > sTics):
        # adjust slope scale
        d = hTics - sTics
        if sBottom == 0:
            sTop += sIntv * d
        else:
            sBottom -= sIntv * (d/2)
            sTop += sIntv * (d/2)
            if d%2 == 1: # odd number
                if (sTop-sMax < sMin-sBottom):
                    sTop += sIntv
                else:
                    sBottom -= sIntv
        sTics = hTics
    if options.verbose:
        print "adjusted H scale:", hBottom, hTop, hIntv, hTics
        print "adjusted S scale:", sBottom, sTop, sIntv, sTics
    return hBottom, hTop, hIntv, sBottom, sTop, sIntv

# print actual graph

def drawCurve(canvas,x,xMin,xMax,y,yMin,yMax,color):
    # calculate derivatives for Bezier points
    x1 = derive(x)
    y1 = derive(y)
    # first point
    coords = [ scaleCanvas (x[0], xMin, xMax, y[0], yMin, yMax) ]
    # remaining points
    for i in range(1,len(x)):
        # bezier segment: first control point
        sx,sy = scaleCanvas (x[i-1] + x1[i-1]/3.0, xMin, xMax,
                             y[i-1] + y1[i-1]/3.0, yMin, yMax)
        coords.append ((round(sx,2), round(sy,2)))
        # second control point
        sx,sy = scaleCanvas (x[i] - x1[i]/3.0, xMin, xMax,
                             y[i] - y1[i]/3.0, yMin, yMax)
        coords.append ((round(sx,2), round(sy,2)))
        # end point
        sx,sy = scaleCanvas (x[i], xMin, xMax, y[i], yMin, yMax)
        coords.append ((round(sx,2), round(sy,2)))
    svgobject.Bezier (canvas, coords, style=svgobject.Style(stroke=color,stroke_width=1.5,fill="none"))

def printCurve(f,x,xMin,xMax,y,yMin,yMax,color):
    # calculate derivatives for Bezier points
    x1 = derive(x)
    y1 = derive(y)
    # begin SVG path statement
    f.write('<path style="stroke:' + color + ';stroke-width:1.5;fill:none" ')
    f.write('d="');
    # first point
    sx,sy = scaleCanvas (x[0], xMin, xMax, y[0], yMin, yMax)
    f.write('M' + str(round(sx,2)) + ',' + str(round(sy,2)))
    # remaining points
    for i in range(1,len(x)):
        # bezier segment: first control point
        sx,sy = scaleCanvas (x[i-1] + x1[i-1]/3.0, xMin, xMax,
                             y[i-1] + y1[i-1]/3.0, yMin, yMax)
        f.write(' C' + str(round(sx,2)) + ',' + str(round(sy,2)))
        # second control point
        sx,sy = scaleCanvas (x[i] - x1[i]/3.0, xMin, xMax,
                             y[i] - y1[i]/3.0, yMin, yMax)
        f.write(',' + str(round(sx,2)) + ',' + str(round(sy,2)))
        # end point
        sx,sy = scaleCanvas (x[i], xMin, xMax, y[i], yMin, yMax)
        f.write(',' + str(round(sx,2)) + ',' + str(round(sy,2)))
    f.write('" />\n')

# scale coordinate such that the interval [x0,x1] is mapped
# to [u0,u1]

def scale(x,x0,x1,u0,u1):
    return u0+(u1-u0)*float(x-x0)/(x1-x0)

# scale point to fit SVG canvas (0,0)-(500,200)

def scaleCanvas(x,x0,x1,y,y0,y1):
    sx = scale(x,x0,x1,0,500)
    sy = scale(y,y0,y1,200,0)
    return sx,sy

# approximate derivative of an array of data points
# using quadratic interpolation between consecutive points

def derive(d):
    n = len(d) # number of points
    if n <= 1:
        return d
    elif n == 2:
        return [d[1]-d[0], d[1]-d[0]]
    else:
        d1 = array.array('d') # result

        # calculate d1[0] from quadratic interpolation of d[0..2]
        d1.append((4*d[1] - 3*d[0] - d[2]) / 2.0)
        for i in range(1, n-1):
            # calculate d1[i] from adjacent points
            if (d[i+1] - d[i] >= 0 and d[i]-d[i-1] <= 0) \
               or (d[i+1] - d[i] <= 0 and d[i]-d[i-1] >= 0):
                di = 0.0
            else:
                di = (d[i+1] - d[i-1]) / 2.0
            d1.append(di)
        # calculate d1[n-1] from quadratic interpolation of d[n-3..n-1]
        d1.append((3*d[n-1] + d[n-3] - 4*d[n-2]) / 2.0)
        return d1

# create web color map (216 colors) as temporary PNM file
# and return the file name
def create_web_color_map ():
    n = 6 # color depth per channel
    width = n * n
    height = n
    fd,fname = tempfile.mkstemp(".pnm")
    fp = os.fdopen (fd, "w")
    fp.write ("P3\n")
    fp.write ("%d %d " % (width, height))
    fp.write ("%d " % (n-1))
    for r in range(n):
        for g in range(n):
            for b in range(n):
                if g>0 or b>0:
                    fp.write (" ")
                fp.write ("%d %d %d" % (r,g,b))
        fp.write ("\n")
    fp.close()
    return fname

main()
